import time
import asyncio

from coderz import Robot, Server


loop = asyncio.get_event_loop()


def wait_fnc(time_to_wait):
    i = 0
    while i < time_to_wait:
        loop.run_until_complete(asyncio.sleep(1))
        i += 1


def ws_move_robot():

    conf = {
        "name": "Me",
        "parts": [
            {"name": "csl", "type": "color"},
            {"name": "csr", "type": "color"},
            {"name": "us", "type": "ultrasonic"},
            {"name": "cs", "type": "controlSystem"},
            {"name": "gyro", "type": "gyro"}
        ],
        "communication": {
            "communication_manager_type": "websocket",
            "server_url": 'http://localhost:1337',
            "socket_emit_route": 'send to vehicle',
            "socket_on_route": 'recieve data',
            "use_authentication_token": False,
        },
        "wait_for_game_start": False
    }

    Server.start()

    robot = Server.get_robot()

    print("------------------------------")

    robot.cs.set_speed(50, 50)
    wait_fnc(5)
    print('robot stop both')
    robot.cs.stop('both')
    wait_fnc(2)
    print('robot set_speed -10, 10')
    robot.cs.set_speed(-10, 10)
    wait_fnc(1)
    print('robot set_speed 50, 50')
    robot.cs.set_speed(50, 50)
    wait_fnc(5)
    print('robot stop both')
    robot.cs.stop('both')
    print('done!')

    Server.stop()

    print('after stop')



from .utils import camel_case_to_snake_case, format_type_to_python
import json
from threading import Event, Thread
import os
from .communication_managers import generate_communication_manager


class Robot:
    ''' Generate a robot object with an api for each of the robot's parts. '''

    def __init__(self, configuration):
        # keep a copy of the configuration for later use.

        ready_event = Event()

        self.__communication_manager = generate_communication_manager(
            configuration["communication"], ready_event)

        self.thread = Thread(target=self.__communication_manager.start)
        self.thread.start()

        ready_event.wait()
        
        self.__configuration = self.__communication_manager.get_configuration()

        # if it is required to wait for a green light from the server in order to run the code, wait.
        if "wait_for_game_start" in self.__configuration and self.__configuration["wait_for_game_start"]:
            self.__communication_manager.wait_for_game_start_message()

        robot_name = self.__configuration["Name"]

        # for each robot-part specified in the configuration, generate an api to it accessible via it's chosen name.
        for part_conf in self.__configuration["Subroutes"]:
            part_name = part_conf['Name']
            methods = part_conf['Methods']

            setattr(self, part_name, Part(
                self.__communication_manager.send_request, part_name, methods, robot_name))

    def get_thread(self):
        return self.thread

    def get_comm_mngr(self):
        return self.__communication_manager

    def print_manual(self):
        ''' Print all the relevant information regarding the robot. '''

        print("Robot name: {0}\n".format(self.__configuration["Name"]))

        print("API options:")

        for robot_part in self.__configuration["Subroutes"]:
            ''' General info about the robot part. '''
            print(F'\nPart Name: {robot_part["Name"]}')
            print("\tUsage:")

            ''' for each of the part methods, print how to use it including all arguments and their types. '''
            for method in robot_part['Methods']:
                method_name = method['Name']
                method_spec = method['Parameters']

                method_arguments = ', '.join("{0} <{1}>".format(
                    argument['Name'], format_type_to_python(argument['Type'])) for argument in method_spec)

                print("\t\trobot.{0}.{1}({2})".format(
                    robot_part['Name'],
                    camel_case_to_snake_case(method_name),
                    method_arguments)
                )


class Part:
    def __init__(self, __send_request_func, part_name, part_methods, robot_name):
        ''' A function to send requests to the robot. '''
        self.__send_request_func = __send_request_func

        # generate a function for each of the robot-part available api calls.
        for method in part_methods:
            method_name = method['Name']
            method_spec = method['Parameters']
            return_type = format_type_to_python(method['ReturnType'])

            method_to_mount = self.__generate_method_to_mount(
                part_name, method_name, method_spec, return_type, robot_name)

            setattr(self, camel_case_to_snake_case(
                method_name), method_to_mount)

    # a function to build a specific api method for a robot-part.
    def __generate_method_to_mount(self, part_name, method_name, method_spec, return_type, robot_name):

        # the method.
        def method_to_mount(*args):
            # make sure the amount of arguments given are correct.
            # TODO: type-check the arguments as well.

            assert len(args) == len(method_spec)

            # which data to send with the api request.
            request_object = {
                "api": part_name,
                "methodName": method_name,
                "parameters": args,
                "playerName": robot_name
            }

            res = self.__send_request_func(
                request_object, return_type)
            return res

        return method_to_mount





# convert javascript style variable names to python style.
# the suggested solution uses re which fails to work on the javascript python compiler:
# **************************************************************
# import re
# def camel_case_to_snake_case(name):
#     s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
#     return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()
# **************************************************************
# hence I cooked up this magnificent pythonic alternative:
def camel_case_to_snake_case(name: str) -> str:
    return "".join(["_" + letter.lower() if letter.isupper() else letter for letter in name]).lstrip('_')


robot_types = {
    "Single": "float",
    "Boolean": "boolean",
    "Vector3": "vector3",
    "Void": None,
    "Single[]": "float[]",
    "Int32": "int",
    "String": "string",
    "Color32": "float[]",
}


def format_type_to_python(unity_type: str) -> str:
    return robot_types[unity_type]



def generate_communication_manager(configuration, ready=None):
    if configuration["communication_manager_type"] == "socket":
        from .socket_communication_manager import SocketCommunicationManager
        return SocketCommunicationManager(configuration)
    elif configuration["communication_manager_type"] == "ide":
        from .ide_communication_manager import WebsocketCommunicationManager
        return WebsocketCommunicationManager(configuration, ready)
    else:
        print("communication manager type {0} not found".format(
            configuration["communication_manager_type"]))



import asyncio
import socketio

# hack into python's low level event loop to inherit its asynchronous black magic abilities.
loop = asyncio.get_event_loop()


class SocketCommunicationManager:
    # this manager allows communication for a 3rd part IDE with either the unity editor or the CoderZ website.

    def __init__(self, configuration):
        # save a copy of the configuration for later use.
        self.__configuration = configuration

        # asyncio black magic variables to be able to work asynchronously while
        # the end-user gets a synchronous experience.
        self.__event = asyncio.Event()
        self.__event.set()
        self.__socket_response = None

        # create a socket connection to the robot.
        self.__sio = socketio.AsyncClient()
        loop.run_until_complete(
            self.__sio.connect(configuration["server_url"]))

        # if a token authentication is required, authenticate.
        if configuration["use_authentication_token"]:
            loop.run_until_complete(self.__authenticate_with_token())

        # a function that is triggered whenever there is a response from the robot.
        @self.__sio.on(self.__configuration["socket_on_route"])
        def on_message(data):
            # when receiving a message, get the response
            # data and set the event to release the waiting on the request sending function.
            self.__socket_response = data
            self.__event.set()

    # general request function to communicate with the robot.
    def send_request(self, request_object, should_wait_for_answer):
        # if the method returns value:
        if should_wait_for_answer is not None:
            res = loop.run_until_complete(
                self.__send_request_and_wait_for_response(request_object))
            return res["result"]
        # if it doesn't return a value:
        else:
            loop.run_until_complete(self.__send_command(request_object))
            return None

    async def __authenticate_with_token(self):
        # validate the socket io connection to the server using a token.

        # ask for a token string from the user.
        token = input(
            "Please insert the token generated by the CoderZ website: ")

        # an event object for async usage
        event = asyncio.Event()

        # prepare a catch function for a response from the server regarding the status of the authentication.
        @self.__sio.on('token validation')
        def on_message(data):
            if data["status"]:
                event.set()
            else:
                raise Exception('token authentication failed')

        # send an authentication request.
        await self.__sio.emit('authenticate IDE', data={"token": token})
        # wait for a response.
        await event.wait()

    async def wait_for_game_start_message(self):
        # wait until the game start before running any code

        # an event object for async usage
        event = asyncio.Event()

        # prepare a catch function for a game ready message from the server.
        @self.__sio.on('run code IDE')
        def on_message():
            event.set()

        # wait for a response.
        await event.wait()

    # send a request for data from the robot.
    async def __send_request_and_wait_for_response(self, request_object):
        # clear the event variable to open up for later responses.
        self.__event.clear()

        # emit the request and wait for a response(A.K.A a self.__event.set() somewhere in the code)
        await self.__sio.emit(self.__configuration["socket_emit_route"], data=request_object)
        await self.__event.wait()

        # return the response.
        return self.__socket_response

    # send a command to the robot which isn't supposed to return data
    async def __send_command(self, request_object):
        # emit the command.
        await self.__sio.emit(self.__configuration["socket_emit_route"], data=request_object)
        # for some weird reason, we need to wait an arbitrarily
        # small amount of time in order for the emit to work.
        # TODO: figure out why...
        await asyncio.sleep(0.000001)

    def get_configurations(self):
        return self.__configuration



from .move_robot import move_robot
from .read_sensor_values import read_sensor_values
from .ws_move_robot import ws_move_robot

example_dict = {
    "move robot": move_robot,
    "read sensor values": read_sensor_values,
    "ws move robot": ws_move_robot,
}



import time
from coderz import Robot


def move_robot():

    conf = {
        "name": "Me",
        "parts": [
            {"name": "csl", "type": "color"},
            {"name": "csr", "type": "color"},
            {"name": "us", "type": "ultrasonic"},
            {"name": "cs", "type": "controlSystem"},
            {"name": "gyro", "type": "gyro"}
        ],
        "communication": {
            "communication_manager_type": "socket",
            "server_url": 'http://localhost:1337',
            "socket_emit_route": 'send to vehicle',
            "socket_on_route": 'recieve data',
            "use_authentication_token": False,
        },
        "wait_for_game_start": False
    }

    robot = Robot(conf)

    print("------------------------------")

    robot.cs.set_speed(50, 50)
    time.sleep(5)
    robot.cs.stop('both')
    time.sleep(2)
    robot.cs.set_speed(-10, 10)
    time.sleep(0.5)
    robot.cs.set_speed(50, 50)
    time.sleep(5)
    robot.cs.stop('both')


from .robot import Robot

conf = {
    "communication": {
        "communication_manager_type": "ide"
    }
}

class Server():
    _is_server_started = False

    _robot = None

    @staticmethod
    def start():
        if Server._is_server_started:
            print('Server already started')
            return

        Server._robot = Robot(conf)
        Server._is_server_started = True
        Server._robot.print_manual()

    @staticmethod
    def stop():
        if not Server._is_server_started:
            print('Not connected')
            return

        comm_mngr = Server._robot.get_comm_mngr()

        comm_mngr.stop()

        Server._robot.get_thread().join()

        Server._is_server_started = False
        Server._robot = None

        print('Server terminated')

    @staticmethod
    def get_robot():
        return Server._robot



import pathlib
import asyncio
import ssl
import time
import json
import websockets
import threading
import sys

ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
localhost_pem = pathlib.Path(__file__).with_name("file.pem")
ssl_context.load_cert_chain(localhost_pem)


SLEEP_TIME = 0.01


class WebsocketCommunicationManager:
    ''' This manager allows communication for a 3rd part IDE with either the unity editor or the CoderZ website. '''

    def __init__(self, configuration, ready=None):
        # save a copy of the configuration for later use.
        self.response_data = None

        self.ready = ready

        self.request_lock = threading.Lock()

        self.response_event = threading.Event()
        self.response_event.set()
        self.command_event = threading.Event()
        self.command_event.set()

        self.request_data = None
        self.response_data = None

        self.interrupt_event = threading.Event()
    
    async def ws_server(self, websocket, path):

        while True:
            if self.interrupt_event.is_set():
                print("Server interrupted")
                break

            await asyncio.sleep(SLEEP_TIME)
            if not self.ready.is_set():
                self.ready.set()

            with self.request_lock:
                if self.request_data != None:
                    await websocket.send(json.dumps(self.request_data))

                    if not self.command_event.is_set():
                        self.command_event.set()

            if not self.response_event.is_set():
                message = await websocket.recv()
                self.wait_response(message)

    def wait_response(self, message):
        jsonLoaded = json.loads(message)

        if jsonLoaded['message'] == 'send data to IDE':
            recieved_data = json.loads(jsonLoaded['data'])
            self.response_data = recieved_data
        elif jsonLoaded['message'] == 'configuration loaded':
            recieved_data = jsonLoaded['data']
            self.response_data = recieved_data
        else:
            print("message not implimented", message)

        self.response_event.set()

    def send_request(self, request_object, should_wait_for_answer):
        ''' General request function to communicate with the robot. '''

        # if the method returns value:
        if should_wait_for_answer is not None:
            res = self.__send_request_and_wait_for_response(request_object)
            return res["result"]
            # if it doesn't return a value:
        else:
            self.__send_command(request_object)
            return None

    '''
    async def wait_for_game_start_message(self):
        # wait until the game start before running any code

        # an event object for async usage
        event = asyncio.Event()

        # prepare a catch function for a game ready message from the server.
        # @self.__sio.on('run code IDE')
        # def on_message():
        #     event.set()

        # wait for a response.
        await event.wait()
    '''

    # send a request for data from the robot.
    def __send_request_and_wait_for_response(self, request_object):
        # clear the event variable to open up for later responses.

        '''
        Emit the request and wait for a response
        (A.K.A a self.__event.set() somewhere in the code)
        '''

        send_obj = {
            'message': 'control vehicle IDE',
            'data': request_object
        }

        self.response_data = None

        with self.request_lock:
            self.request_data = send_obj

        self.response_event.clear()
        self.response_event.wait()

        # return the response.
        return self.response_data

    # send a command to the robot which isn't supposed to return data
    def __send_command(self, request_object):
        # emit the command.
        send_obj = {
            'message': 'control vehicle IDE',
            'data': request_object
        }
        with self.request_lock:
            self.request_data = send_obj
        self.command_event.clear()
        self.command_event.wait()

    def start(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        start_server = websockets.serve(
            self.ws_server, "localhost", 25842, ssl=ssl_context)

        websocket_server = loop.run_until_complete(start_server)
        
        while True:
            loop.run_until_complete(asyncio.sleep(1))
            if self.interrupt_event.is_set():
                break

    def get_configuration(self):
        ''' This is blocking all I/O 🐒 '''
        send_obj = {
            'message': 'load configurations'
        }

        self.response_data = None

        with self.request_lock:
            self.request_data = send_obj

        self.response_event.clear()
        self.response_event.wait()

        # return the response.
        return self.response_data

    def stop(self):
        self.interrupt_event.set()



from coderz import Robot


def read_sensor_values():
    conf = {
        "name": "Me",
        "parts": [
            {"name": "csl", "type": "color"},
            {"name": "csr", "type": "color"},
            {"name": "us", "type": "ultrasonic"},
            {"name": "cs", "type": "controlSystem"},
            {"name": "gyro", "type": "gyro"}
        ],
        "communication": {
            "communication_manager_type": "socket",
            "server_url": 'http://localhost:1337',
            "socket_emit_route": 'send to vehicle',
            "socket_on_route": 'recieve data',
            "use_authentication_token": False
        },
        "wait_for_game_start": False
    }

    robot = Robot(conf)

    robot.print_manual()

    print("------------------------------")

    print("the left color sensor reads: {0}".format(robot.csl.get_color_name()))
    # print("the right color sensor reads: {0}".format(robot.csr.get_color_name()))
    print("the gyro reads: {0}".format(robot.gyro.get_angle_y()))
    print("the ultrasonic reads: {0}".format(robot.us.get_distance()))



import setuptools
#print(setuptools.find_packages(where='src'))
#exit()
with open("Readme.md", "r") as fh:
    long_description = fh.read()
setuptools.setup(
    name='coderz',
    version='0.1.12',
    author="CoderZ",
    author_email="app@gocoderz.com",
    python_requires='>3.0',
    description="CoderZ External IDE Tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/go-coderz/coderz-python",
    install_requires=[
    'websockets>=7.0',
    'python-socketio==4.0.2',
    'aiohttp>=3.0.0'
    ],
    packages=setuptools.find_packages(where='src'),
    package_dir={'': 'src'},
    package_data={'': ['*']},
    include_package_data=True,
    classifiers=[
         "Programming Language :: Python :: 3",
         "License :: OSI Approved :: MIT License",
         "Operating System :: OS Independent",
    ],
)
